package models.actions

import models.actions.AuthenticationRequest

import play.api.mvc.{ActionBuilder, BodyParsers, Request, Result, AnyContent, Cookie}
import play.filters.csrf.CSRF

import javax.inject.Inject
import scala.concurrent.{Future, ExecutionContext}


/**
 * The authentication action builder that combines the request transformation and filtering.
 */
class AuthenticatedUserAction @Inject()(authenticationTransformer: AuthenticationTransformer, authenticationFilter: AuthenticationFilter)
                                       (implicit val parser: BodyParsers.Default, val executionContext: ExecutionContext)
    extends ActionBuilder[AuthenticationRequest, AnyContent] {

    /**
     * Invoke the main controller block, with the transformations and filtering middleware, and the CSRF token injection.
     *
     * @param request The incoming request.
     * @param block The block of code to invoke.
     * @return A future of the result.
     */
    override def invokeBlock[A](request: Request[A], block: AuthenticationRequest[A] => Future[Result]): Future[Result] = {
        val result: Future[Result] = (authenticationTransformer andThen authenticationFilter).invokeBlock(request, block)

        (result).map(_result => {
            // Add CSRF token to response
            val token = CSRF.getToken(request).map(_.value).getOrElse("")
            _result.withHeaders("Csrf-Token" -> token).withCookies(Cookie(name = "PLAY_CSRF_TOKEN", value = token))
        })
    }
}
